/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2011 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */
package org.ow2.weblab.tools.maven;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;

import org.apache.maven.archetype.ArchetypeGenerationRequest;
import org.apache.maven.archetype.ArchetypeGenerationResult;
import org.apache.maven.archetype.ArchetypeManager;
import org.apache.maven.archetype.exception.ArchetypeGenerationConfigurationFailure;
import org.apache.maven.archetype.exception.ArchetypeNotConfigured;
import org.apache.maven.archetype.exception.ArchetypeNotDefined;
import org.apache.maven.archetype.exception.ArchetypeSelectionFailure;
import org.apache.maven.archetype.exception.UnknownArchetype;
import org.apache.maven.archetype.exception.UnknownGroup;
import org.apache.maven.archetype.ui.ArchetypeGenerationConfigurator;
import org.apache.maven.archetype.ui.ArchetypeSelector;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.codehaus.plexus.components.interactivity.PrompterException;
import org.codehaus.plexus.util.StringUtils;
import org.ow2.weblab.tools.maven.util.Convention;
import org.ow2.weblab.tools.maven.util.Message;
import org.ow2.weblab.tools.maven.util.WebLabInterface;

/**
 * This plug-in allow to generate archetype contains WebLab's service(s)
 * 
 * @goal generate
 * @execute phase="generate-sources"
 * @requiresProject false
 */
public class WebLabGenerate extends AbstractMojo {


	/**
	 * The archetype's groupId.
	 * 
	 * @parameter expression="${archetypeGroupId}" default-value="org.ow2.weblab.tools.maven"
	 */
	private String archetypeGroupId;


	/**
	 * The archetype's version.
	 * 
	 * @parameter expression="${archetypeVersion}" default-value="1.0"
	 */
	private String archetypeVersion;


	/**
	 * Service name.
	 * 
	 * @parameter expression="${serviceName}"
	 * @required
	 */
	private String serviceName;


	/**
	 * Interface(s) WebLab.
	 * 
	 * @parameter expression="${webLabInterface}"
	 * @required
	 */
	private String webLabInterface;


	/**
	 * ArtifactId.
	 * 
	 * This parameter is not used in this plugin but compulsory for archetypes.
	 * 
	 * @parameter expression="${artifactId}"
	 * @required
	 */
	private String artifactId;


	/**
	 * GroupId name.
	 * 
	 * @parameter expression="${groupId}"
	 */
	private String groupIdName;


	/**
	 * Package name.
	 * 
	 * @parameter expression="${package}"
	 */
	private String packageName;


	/**
	 * @parameter expression="${basedir}"
	 */
	private File basedir;


	/**
	 * @parameter expression="${session}"
	 * @readonly
	 */
	private MavenSession session;


	// //////////////////////////////////////////////////////////////////////

	/**
	 * Local Maven repository.
	 * 
	 * @parameter expression="${localRepository}" required
	 * @readonly
	 */
	private ArtifactRepository localRepository;


	/**
	 * The archetype's repository.
	 * 
	 * @parameter expression="${archetypeRepository}"
	 */
	private String archetypeRepository;


	/**
	 * List of remote repositories used by the resolver.
	 * 
	 * @parameter expression="${project.remoteArtifactRepositories}"
	 * @readonly
	 */
	private List<ArtifactRepository> remoteArtifactRepositories;


	/**
	 * The archetype's catalogs. It is a comma separated list of catalogs.
	 * Catalogs use scheme:
	 * <ul>
	 * <li>'<code>file://...</code>' with <code>archetype-catalog.xml</code> automatically appended when pointing to a directory</li>
	 * <li>'<code>http://...</code>' with <code>archetype-catalog.xml</code> always appended</li>
	 * <li>'<code>local</code>' which is the shortcut for ' <code>file://~/.m2/archetype-catalog.xml</code>'</li>
	 * <li>'<code>remote</code>' which is the shortcut for Maven Central repository, ie '<code>http://repo1.maven.org/maven2</code>'</li>
	 * <li>'<code>internal</code>' which is an internal catalog</li>
	 * </ul>
	 * 
	 * Since 2.0-alpha-5, default value is no longer <code>internal,local</code> but <code>remote,local</code>. If Maven Central repository catalog file
	 * is empty, <code>internal</code> catalog is used instead.
	 * 
	 * @parameter expression="${archetypeCatalog}" default-value="internal"
	 * 
	 */
	private String archetypeCatalog;


	/** @component */
	private ArchetypeManager archetype;


	/** @component */
	private ArchetypeSelector selector;


	/** @component */
	private ArchetypeGenerationConfigurator configurator;


	// ///////////////////////////////////////////////////////////////////////

	/**
	 * The name of archetype which contains resources
	 */
	private static final String ARCHETYPE_ARTIFACT_ID = "weblab-archetype-resources";


	/**
	 * The minimum number of interface to implement before to concatenate the
	 * filename with service name and interface name
	 */
	private static final int MIN_IMPLEMENTATION_TO_CONCATENATE_SERVICE_NAME = 1;


	/**
	 * Directory path of the main classes
	 */
	private static final String DIRECTORY_PATH_MAIN_CLASSES = "/src/main/java/";


	/**
	 * Directory path of the test classes
	 */
	private static final String DIRECTORY_PATH_TEST_CLASSES = "/src/test/java/";


	/**
	 * Extension of java classes
	 */
	private static final String EXT_JAVA_CLASS = ".java";


	/**
	 * Suffix of test class name
	 */
	private static final String SUFF_TEST_CLASS = "Test";


	/**
	 * Character sharp
	 */
	private static final String SHARP = "#";


	/**
	 * Character slash
	 */
	private static final String SLASH = "/";


	/**
	 * Character dot
	 */
	private static final String DOT = ".";


	/**
	 * Character comma
	 */
	private static final String COMMA = ",";


	/**
	 * Error message for service request
	 */
	private static final String ERROR_SERVICE = "Impossible to generated service.";


	/**
	 * Error message for resource request
	 */
	private static final String ERROR_RESOURCE = "Impossible to generated resource.";


	@Override
	public void execute() throws MojoExecutionException, MojoFailureException {
		// TODO Create a catalog with WebLab archetypes
		this.getLog().info("The following message concerning the catalog is not a problem.");

		if (this.artifactId == null) {
			throw new MojoFailureException("ArtifactId must be defined!");
		}

		/*
		 * Checks if WebLab interface(s) informed exist and create a list of these interfaces
		 */
		final Set<WebLabInterface> listInterfacesWebLab = this.webLabInterfaceList(this.webLabInterface);

		/*
		 * Checks service name
		 */
		if (Convention.checkClassName(this.serviceName) == null) {
			throw new MojoFailureException("Service name must be defined!");
		}
		if (Convention.checkClassName(this.serviceName) == Message.ERROR) {
			this.errorClassName();
			throw new MojoFailureException("Invalid class name!");
		} else if (Convention.checkClassName(this.serviceName) == Message.WARN) {
			this.warnClassName();
		}

		/*
		 * Checks if groupId or package respect the Java conventions.
		 * 
		 * Note : if package is not informed then groupId is used such as package
		 */
		if (Convention.checkPackage(this.packageName) != null) {
			if (Convention.checkPackage(this.packageName) == Message.ERROR) {
				this.errorPackage();
				throw new MojoFailureException("Invalid package name!");
			} else if (Convention.checkPackage(this.packageName) == Message.WARN) {
				this.warnPackage();
			}
		} else if (Convention.checkGroupID(this.groupIdName) != null) {
			if (Convention.checkGroupID(this.groupIdName) == Message.ERROR) {
				this.errorPackage();
				throw new MojoFailureException("Invalid package name!");
			} else if (Convention.checkPackage(this.groupIdName) == Message.WARN) {
				this.warnPackage();
			}
		} else {
			this.getLog().debug("Using default groupId for each archetype.");
		}

		/*
		 * Properties used for all artifacts which allows to parsed several resources contains in these artifact
		 */
		final Properties executionProperties = this.session.getExecutionProperties();

		/*
		 * Main request which allows to get main resources contained in artifact WebLabArchetype: calls the weblab-archetype-resource plugin
		 */
		final ArchetypeGenerationRequest requestResource = new ArchetypeGenerationRequest();
		requestResource.setArchetypeGroupId(this.archetypeGroupId);
		requestResource.setArchetypeArtifactId(WebLabGenerate.ARCHETYPE_ARTIFACT_ID);
		requestResource.setArchetypeVersion(this.archetypeVersion);
		requestResource.setOutputDirectory(this.basedir.getAbsolutePath());
		requestResource.setLocalRepository(this.localRepository);
		requestResource.setArchetypeRepository(this.archetypeRepository);
		requestResource.setRemoteArtifactRepositories(this.remoteArtifactRepositories);

		/*
		 * List of classes names associate at interface name
		 */
		final Map<String, String> classNames = new HashMap<String, String>();

		/*
		 * The endPoint(s) to inserted in cxf-servlet.xml file
		 */
		final StringBuilder buffer = new StringBuilder();

		/*
		 * Build request
		 */
		final ArchetypeGenerationRequest requestService = new ArchetypeGenerationRequest();
		requestService.setArchetypeGroupId(this.archetypeGroupId);
		requestService.setArchetypeVersion(this.archetypeVersion);
		requestService.setOutputDirectory(this.basedir.getAbsolutePath());
		requestService.setLocalRepository(this.localRepository);
		requestService.setArchetypeRepository(this.archetypeRepository);
		requestService.setRemoteArtifactRepositories(this.remoteArtifactRepositories);

		/*
		 * Generation of WebLab service(s)
		 */
		for (final WebLabInterface webLabInterfaces : listInterfacesWebLab) {

			/*
			 * Recover WebLab interface name
			 */
			final String webLabInterfaceName = webLabInterfaces.name().toLowerCase();

			/*
			 * Generate class name with WebLab interface
			 */
			final String className = this.generateClassName(listInterfacesWebLab.size(), webLabInterfaces, this.serviceName);
			classNames.put(className, webLabInterfaceName);

			/*
			 * Generate test class name with class name
			 */
			final String classNameTest = this.generateClassNameTest(className);

			/*
			 * Add properties
			 */
			executionProperties.setProperty("className", className);
			executionProperties.setProperty("classNameTest", classNameTest);

			requestService.setArchetypeArtifactId("weblab-archetype-" + webLabInterfaceName);


			try {

				try {
					// Selects archetype with request
					this.selector.selectArchetype(requestService, Boolean.FALSE, this.archetypeCatalog);
				} catch (final UnknownGroup ug) {
					throw new MojoFailureException(WebLabGenerate.ERROR_SERVICE + requestService.getArchetypeArtifactId(), ug);
				} catch (final ArchetypeSelectionFailure asf) {
					throw new MojoFailureException(WebLabGenerate.ERROR_SERVICE + requestService.getArchetypeArtifactId(), asf);
				}

				try {
					// Configures archetype with request
					this.configurator.configureArchetype(requestService, Boolean.FALSE, executionProperties);
				} catch (final ArchetypeNotConfigured anc) {
					throw new MojoFailureException(WebLabGenerate.ERROR_SERVICE + requestService.getArchetypeArtifactId(), anc);
				} catch (final ArchetypeGenerationConfigurationFailure agcf) {
					throw new MojoFailureException(WebLabGenerate.ERROR_SERVICE + requestService.getArchetypeArtifactId(), agcf);
				}

			} catch (final ArchetypeNotDefined and) {
				throw new MojoFailureException(WebLabGenerate.ERROR_SERVICE + requestService.getArchetypeArtifactId(), and);
			} catch (final UnknownArchetype ua) {
				throw new MojoFailureException(WebLabGenerate.ERROR_SERVICE + requestService.getArchetypeArtifactId(), ua);
			} catch (final IOException ioe) {
				throw new MojoFailureException(WebLabGenerate.ERROR_SERVICE + requestService.getArchetypeArtifactId(), ioe);
			} catch (final PrompterException pe) {
				throw new MojoFailureException(WebLabGenerate.ERROR_SERVICE + requestService.getArchetypeArtifactId(), pe);
			}

			/*
			 * Generates partial artifact
			 */
			final ArchetypeGenerationResult generationResult = this.archetype.generateProjectFromArchetype(requestService);

			if (generationResult.getCause() != null) {
				throw new MojoFailureException(generationResult.getCause(), generationResult.getCause().getMessage(), generationResult.getCause().getMessage());
			}

			/*
			 * Creates end point for framework CXF
			 */
			buffer.append("<jaxws:endpoint id=\"" + webLabInterfaceName + "\" implementor=\"" + WebLabGenerate.SHARP + className + "\" address=\""
					+ WebLabGenerate.SLASH + webLabInterfaceName + "\" />\n\t");
			buffer.append("<bean class=\"" + requestService.getPackage() + "." + className + "\" id=\"" + className + "\" />\n\t");
		}

		/*
		 * Transforms package in path
		 */
		final String pathPackage = this.packageToPath(requestService.getPackage());

		/*
		 * Adds a end point property and updates request resource with set of properties
		 */
		executionProperties.setProperty("endPoint", buffer.toString());
		requestResource.setProperties(executionProperties);

		/*
		 * Generation of resources
		 */
		try {
			try {
				this.selector.selectArchetype(requestResource, Boolean.FALSE, this.archetypeCatalog);
			} catch (final UnknownGroup ug) {
				throw new MojoFailureException(WebLabGenerate.ERROR_RESOURCE + requestService.getArchetypeArtifactId(), ug);
			} catch (final ArchetypeSelectionFailure asf) {
				throw new MojoFailureException(WebLabGenerate.ERROR_RESOURCE + requestService.getArchetypeArtifactId(), asf);
			}

			try {
				this.configurator.configureArchetype(requestResource, Boolean.FALSE, executionProperties);
			} catch (final ArchetypeNotConfigured anc) {
				throw new MojoFailureException(WebLabGenerate.ERROR_RESOURCE + requestService.getArchetypeArtifactId(), anc);
			} catch (final ArchetypeGenerationConfigurationFailure agcf) {
				throw new MojoFailureException(WebLabGenerate.ERROR_RESOURCE + requestService.getArchetypeArtifactId(), agcf);
			}
		} catch (final ArchetypeNotDefined and) {
			throw new MojoFailureException(WebLabGenerate.ERROR_RESOURCE + requestService.getArchetypeArtifactId(), and);
		} catch (final UnknownArchetype ua) {
			throw new MojoFailureException(WebLabGenerate.ERROR_RESOURCE + requestService.getArchetypeArtifactId(), ua);
		} catch (final IOException ioe) {
			throw new MojoFailureException(WebLabGenerate.ERROR_RESOURCE + requestService.getArchetypeArtifactId(), ioe);
		} catch (final PrompterException pe) {
			throw new MojoFailureException(WebLabGenerate.ERROR_RESOURCE + requestService.getArchetypeArtifactId(), pe);
		}

		/*
		 * End artifact generation
		 */
		final ArchetypeGenerationResult generationResult = this.archetype.generateProjectFromArchetype(requestResource);
		if (generationResult.getCause() != null) {
			throw new MojoFailureException(generationResult.getCause(), generationResult.getCause().getMessage(), generationResult.getCause().getMessage());
		}

		/*
		 * Renames files
		 */
		final String theArtifactId = requestResource.getArtifactId();
		final File projectBasedir = new File(this.basedir, theArtifactId);
		this.renameFiles(projectBasedir, pathPackage, classNames);
	}


	/*
	 * Error message for interface WebLab parameter
	 */
	private void errorWebLabInterface() {
		this.getLog().error(""); // keep [INFO] at the beginning of line
		this.getLog().error("You can use only this interfaces:");
		this.getLog().error(""); // keep [INFO] at the beginning of line
		for (final WebLabInterface interfaces : WebLabInterface.values()) {
			this.getLog().error(interfaces.ordinal() + 1 + " : " + interfaces.name().toLowerCase());
		}
		this.getLog().error(""); // keep [INFO] at the beginning of line
	}


	/*
	 * Error message for package parameter
	 */
	private void errorPackage() {
		this.getLog().error(""); // keep [INFO] at the beginning of line
		this.getLog().error("Invalid package name. A package name has to:");
		this.getLog().error(""); // keep [INFO] at the beginning of line
		this.getLog().error("be all lower case");
		this.getLog().error("used only [a-z] [0-9] and '.'");
		this.getLog().error("not start with [0-9] or '.'");
		this.getLog().error("not end with [0-9] or '.'");
		this.getLog().error(""); // keep [INFO] at the beginning of line
		this.getLog().error("Note: if package is not informed then groupId is used such as package");
		this.getLog().error(""); // keep [INFO] at the beginning of line
	}


	/*
	 * Warning message for package parameter
	 */
	private void warnPackage() {
		this.getLog().warn(""); // keep [INFO] at the beginning of line
		this.getLog().warn("Discouraged package name. By convention, package names have to:");
		this.getLog().warn(""); // keep [INFO] at the beginning of line
		this.getLog().warn("be all lower case");
		this.getLog().warn("used only [a-z] [0-9] and '.'");
		this.getLog().warn("not start with [0-9] or '.'");
		this.getLog().warn("not end with [0-9] or '.'");
		this.getLog().warn(""); // keep [INFO] at the beginning of line
		this.getLog().warn("Note: if package is not informed then groupId is used such as package");
		this.getLog().warn(""); // keep [INFO] at the beginning of line
	}


	/*
	 * Error message for class name parameter
	 */
	private void errorClassName() {
		this.getLog().error(""); // keep [INFO] at the beginning of line
		this.getLog().error("Invalid class name. A class name has to:");
		this.getLog().error(""); // keep [INFO] at the beginning of line
		this.getLog().error("start with upper case");
		this.getLog().error("used only [a-z] [A-Z] and [0-9]");
		this.getLog().error("first letter of each word capitalized");
		this.getLog().error(""); // keep [INFO] at the beginning of line
		this.getLog().error("Note: service name is used to generate class(es) name(s)");
		this.getLog().error(""); // keep [INFO] at the beginning of line

	}


	/*
	 * Warning message for class name parameter
	 */
	private void warnClassName() {
		this.getLog().warn(""); // keep [INFO] at the beginning of line
		this.getLog().warn("Discouraged class name. By convention, class names have to:");
		this.getLog().warn(""); // keep [INFO] at the beginning of line
		this.getLog().warn("start with upper case");
		this.getLog().warn("used only [a-z] [A-Z] and [0-9]");
		this.getLog().warn("first letter of each word capitalized");
		this.getLog().warn(""); // keep [INFO] at the beginning of line
		this.getLog().warn("Note: service name is used to generate class(es) name(s)");
		this.getLog().warn(""); // keep [INFO] at the beginning of line

	}


	/*
	 * Change package in path
	 */
	private String packageToPath(final String packageService) {
		final StringBuilder buffer = new StringBuilder();
		buffer.append(packageService.replace(WebLabGenerate.DOT, WebLabGenerate.SLASH));
		buffer.append(WebLabGenerate.SLASH);
		return buffer.toString();
	}


	/*
	 * Rename files of generated project
	 */
	private void renameFiles(final File projectBasedir, final String pathPackage, final Map<String, String> classNames) throws MojoFailureException {
		/*
		 * Classes of the project
		 */
		final File directoryClasses = new File(projectBasedir + WebLabGenerate.DIRECTORY_PATH_MAIN_CLASSES + pathPackage);
		if (!directoryClasses.isDirectory()) {
			throw new MojoFailureException(directoryClasses.getAbsolutePath() + "is not a directory !");
		}
		for (final File classes : directoryClasses.listFiles()) {
			for (final Entry<String, String> entry : classNames.entrySet()) {
				if (classes.getAbsolutePath().contains(entry.getValue())) {
					final StringBuilder renameClass = new StringBuilder();
					renameClass.append(classes.getAbsolutePath().replace(classes.getName(), ""));
					renameClass.append(entry.getKey() + WebLabGenerate.EXT_JAVA_CLASS);
					classes.renameTo(new File(renameClass.toString()));
				}
			}
		}

		/*
		 * Test classes of the project
		 */
		final File directoryTestClasses = new File(projectBasedir + WebLabGenerate.DIRECTORY_PATH_TEST_CLASSES + pathPackage);
		if (!directoryTestClasses.isDirectory()) {
			throw new MojoFailureException(directoryTestClasses.getAbsolutePath() + "is not a directory !");
		}

		for (final File testClasses : directoryTestClasses.listFiles()) {
			for (final Entry<String, String> entry : classNames.entrySet()) {
				if (testClasses.getAbsolutePath().contains(entry.getValue())) {
					final String className = this.generateClassNameTest(entry.getKey());
					final StringBuilder renameTestClass = new StringBuilder();
					renameTestClass.append(testClasses.getAbsolutePath().replace(testClasses.getName(), ""));
					renameTestClass.append(className + WebLabGenerate.EXT_JAVA_CLASS);
					testClasses.renameTo(new File(renameTestClass.toString()));
				}
			}
		}
	}


	/*
	 * Manages WebLab interfaces and combines names with numbers
	 */
	private Set<WebLabInterface> webLabInterfaceList(final String webLabInterfaceInput) throws MojoFailureException {
		final String interfacesList[] = webLabInterfaceInput.split(WebLabGenerate.COMMA);
		final Set<WebLabInterface> listInterfacesWebLab = new HashSet<WebLabInterface>();
		for (final String s : interfacesList) {
			final WebLabInterface interfaceGiveByUser;
			if (StringUtils.isNumeric(s)) {
				final int i;
				try {
					i = Integer.parseInt(s);
				} catch (final NumberFormatException e) {
					this.errorWebLabInterface();
					throw new MojoFailureException("The number must be an integer!", e);
				}

				/*
				 * Test if number is correct, it must be between one and the total number of interfaces
				 */
				if ((i > 0) && (i <= WebLabInterface.values().length)) {
					interfaceGiveByUser = WebLabInterface.values()[i - 1];
				} else {
					this.errorWebLabInterface();
					throw new MojoFailureException("The number does not match with a WebLab interface!");
				}

			} else {
				try {
					final String enumWebLabInterface = s.toUpperCase();
					interfaceGiveByUser = Enum.valueOf(WebLabInterface.class, enumWebLabInterface);
				} catch (final IllegalArgumentException iae) {
					this.errorWebLabInterface();
					throw new MojoFailureException("Interface " + s + " does not exist!", iae);
				}
			}

			/*
			 * Fail if an interface is added twice
			 */
			if (!listInterfacesWebLab.add(interfaceGiveByUser)) {
				throw new MojoFailureException(StringUtils.capitalizeFirstLetter(interfaceGiveByUser.name().toLowerCase()) + " is used twice!");
			}
		}

		return listInterfacesWebLab;
	}


	/*
	 * Return the name of generated class from service name
	 */
	private String generateClassName(final int numberService, final WebLabInterface theWebLabInterface, final String theServiceName) {
		/*
		 * If there are many services to implement, so the classes generated must concatenate the service name given by user with the WebLab interface name
		 */
		final String className;
		if (numberService > WebLabGenerate.MIN_IMPLEMENTATION_TO_CONCATENATE_SERVICE_NAME) {

			className = theServiceName + StringUtils.capitalizeFirstLetter(theWebLabInterface.toString().toLowerCase());
			/*
			 * Else the filename must be the service name
			 */
		} else {
			className = theServiceName;
		}
		return className;
	}


	/*
	 * Return the name of generated test class
	 */
	private String generateClassNameTest(final String className) {
		return className + WebLabGenerate.SUFF_TEST_CLASS;
	}

}